/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include "fist.h"
#include "common.h"
#include "ad_aggregate.h"
#include "stream_header.h"
#include "serializer.h"
#include <linux/utsname.h>
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif

#define OP_PROFILING

/* 32 << 32 cycles should be enough even for VERY slow FSs :-) */
#define MAX_DISTR_DIGIT 32
/* XXX: Should be a dynamic parameter */
/* set to 0 to have no timeline and just the totals */
#define MAX_DISTR_TIME 1000

/* This whole structure is smaller than 4K if MAX_DISTR_TIME = 0 */
struct aggregate_pd {
	unsigned long opcounts[OP_MAX];
#ifdef OP_PROFILING
	unsigned long long init_cycle;
	unsigned long long tot_cycles[OP_MAX];
	unsigned long distribution[OP_MAX][MAX_DISTR_TIME + 1][MAX_DISTR_DIGIT + 1];
#endif
};

#define pd(ptr) (((struct aggregate_pd *)(ptr)->ad_private_data))

struct aggregate_context {
#ifdef OP_PROFILING
  msg_seqnum_t seqnum;
    unsigned long long init_cycle;
#endif
};



#ifdef OP_PROFILING

#define INIT_CYCLE (((struct aggregate_context*)*context)->init_cycle)

#define RDTSC(qp) \
    do { \
        unsigned long lowPart, highPart; \
        __asm__ __volatile__("rdtsc" : "=a" (lowPart), "=d" (highPart)); \
        qp = (((unsigned long long) highPart) << 32) | lowPart; \
    } while (0);

#define ALLOC_CONTEXT *context = (void*)KMALLOC(sizeof(struct aggregate_context), GFP_KERNEL);
#define SET_CONTEXT_SEQNUM (*(struct aggregate_context**)context)->seqnum = atomic_read(&sequence_number);\
                           atomic_inc(&sequence_number);
#define WRITE_CONTEXT_SEQNUM memcpy(buf->data+SEQNUM_LOC,&(*(struct aggregate_context**)context)->seqnum,sizeof(msg_seqnum_t));
#define FREE_CONTEXT KFREE(*context);
#define INIT_CONTEXT RDTSC(INIT_CYCLE);

#define UPDATE_COUNTER(op) \
    do { \
	unsigned long long l; \
	unsigned long long delay; \
	unsigned long long ll; \
	int i, ii; \
	RDTSC(l); \
        delay = (l - INIT_CYCLE); \
	pd(this)->tot_cycles[(op)] += delay; \
	ll = 32L; \
	for (i = 0; i < MAX_DISTR_DIGIT; i++) { \
	    if (delay < ll) \
		break; \
	    ll = ll << 1; \
	} \
	ii = (int)((l - pd(this)->init_cycle) >> 32); \
	if (ii > MAX_DISTR_TIME) \
	    ii = MAX_DISTR_TIME; \
	pd(this)->distribution[(op)][ii][i]++; \
    } while(0);

#else

#define ALLOC_CONTEXT
#define SET_CONTEXT_SEQNUM
#define WRITE_CONTEXT_SEQNUM
#define FREE_CONTEXT
#define INIT_CONTEXT
#define UPDATE_COUNTER(op)

#endif

#define PRE_HEADER(op) \
	struct buffer *buf; \
	pd(this)->opcounts[(op)]++; \
        buf = get_buffer(); \
        ASSERT(buf); \
	ALLOC_CONTEXT; \
        SET_CONTEXT_SEQNUM; \
        WRITE_CONTEXT_SEQNUM;

#define PRE_FOOTER \
	INIT_CONTEXT;

#define POST_HEADER(op) \
        struct buffer *buf; \
	UPDATE_COUNTER(op); \
        buf = get_buffer(); \
        ASSERT(buf); \
        WRITE_CONTEXT_SEQNUM;

#define POST_FOOTER \
	FREE_CONTEXT;

static void add_op(struct assembly_driver *this, struct trace_destination *dest, struct buffer *buf, op_t op)
{
	op_t tmp = op;
	msg_seqnum_t seqnum;
	memcpy(buf->data, &tmp, sizeof(tmp));

	//Write Sequence Number to buffer
	seqnum = atomic_read(&sequence_number);
	atomic_inc(&sequence_number);
	memcpy(buf->data+SEQNUM_LOC,&seqnum,sizeof(seqnum));
}

void aggregate_llseek_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, loff_t offset, int origin, void **context)
{
    PRE_HEADER(OP_LLSEEK);
    PRE_FOOTER;
}

void aggregate_llseek_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, loff_t offset, int origin, loff_t ret, void **context)
{
    POST_HEADER(OP_LLSEEK);
    POST_FOOTER;
}

void aggregate_open_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode, struct file *file, void **context)
{
    PRE_HEADER(OP_OPEN);
    PRE_FOOTER;
}

void aggregate_open_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode, struct file *file, int ret, void **context)
{
    POST_HEADER(OP_OPEN);
    POST_FOOTER;
}

static int aggregate_init(struct assembly_driver *this, void *options)
{
    int err = -ENOMEM;
    ASSERT(this);
    ASSERT(this->ad_type == AD_AGGREGATE);

#if (MAX_DISTR_TIME > 0)
    this->ad_private_data = vmalloc(sizeof(struct aggregate_pd));
#else
    this->ad_private_data = KMALLOC(sizeof(struct aggregate_pd), GFP_KERNEL);
#endif
    if (!this->ad_private_data)
	goto out;

    memset(pd(this), 0, sizeof(struct aggregate_pd));
#ifdef OP_PROFILING
    RDTSC(pd(this)->init_cycle);
#endif
    sema_init(&this->ad_odriver_sem, 1);
    err = 0;

  out:
    return err;

}

static void aggregate_release(struct assembly_driver *this)
{
    ASSERT(this);
    ASSERT(this->ad_type == AD_AGGREGATE);
    if (this->ad_private_data) {
#if (MAX_DISTR_TIME > 0)
	vfree(this->ad_private_data);
#else
	KFREE(this->ad_private_data);
#endif
	this->ad_private_data = NULL;
    }
}

static inline void aggregate_add_header(struct assembly_driver *this, struct trace_destination *dest)
{
    struct stream_header header;
    header.magic = htonl(TRACEFS_STREAM_MAGIC);
    header.version = htonl(VERSION_CODE);
    down(&this->ad_odriver_sem);
    if (dest->td_state != TD_ERROR) {
	if (dest->td_od->od_ops->write(dest->td_od, &header, sizeof(header),OO_KERNEL_BUFFER) != sizeof(header))
	    dest->td_state = TD_ERROR;
    }
    up(&this->ad_odriver_sem);
}

void aggregate_start_trace(struct assembly_driver *this, struct trace_destination *dest)
{
    struct buffer *buf = get_buffer();
    op_t tmp = MSG_START;
    aggregate_add_header(this, dest);
    if (!buf)
	goto out;
    /*
     * We are not using the standard start_pre_trace_message() because
     * we don't want to have the MSG_PNAME message being created for this
     * trace message.
     */

    memcpy(buf->data, &tmp, sizeof(tmp));

    //Set Sequence number for start message to 0
    memset(buf->data+SEQNUM_LOC,0,sizeof(msg_seqnum_t));
    //Initialize sequence_number = 1
    atomic_set(&sequence_number,1);

    add_tv(buf, ARG_PID, current->pid);
    add_tv(buf, ARG_UID, current->euid);
    add_tv(buf, ARG_GID, current->egid);
    /* uname stuff */
    add_tlv(buf, ARG_SYSNAME, sizeof(system_utsname.sysname), system_utsname.sysname);
    add_tlv(buf, ARG_RELEASE, sizeof(system_utsname.release), system_utsname.release);
    add_tlv(buf, ARG_VERSION, sizeof(system_utsname.version), system_utsname.version);
    add_tlv(buf, ARG_MACHINE, sizeof(system_utsname.machine), system_utsname.machine);

    end_trace_message(buf);
    down(&this->ad_odriver_sem);
    if (dest->td_state != TD_ERROR) {
	if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data,OO_KERNEL_BUFFER) != buf->ptr - buf->data)
		dest->td_state = TD_ERROR;
    }
    up(&this->ad_odriver_sem);
    release_buffer(buf);
  out:
    return;
}

#ifdef OP_PROFILING
void aggregate_stop_trace(struct assembly_driver *this, struct trace_destination *dest)
{
	struct buffer *buf = get_buffer();
	field_t i, ii, last_tick;
	unsigned long long now;

	RDTSC(now);
	now -= pd(this)->init_cycle;
	last_tick = (field_t)(now >> 32);
	if (last_tick > MAX_DISTR_TIME)
	    last_tick = MAX_DISTR_TIME;

	if (!buf)
	    goto out;
	for (i = 0; i < OP_MAX; i++) {
	    if (pd(this)->opcounts[i] > 0) {
		add_op(this, dest, buf, pre_op((i)));
		add_tv(buf, ARG_MSG_ID, i);
		add_tv(buf, ARG_ULONG, pd(this)->opcounts[i]);
		add_tv(buf, ARG_UULONG, pd(this)->tot_cycles[i]);
		for (ii = 0; ii < last_tick + 1; ii++) {
		    add_tlv(buf, ARG_DATA, (MAX_DISTR_DIGIT + 1) * sizeof(unsigned long) , &pd(this)->distribution[i][ii][0]);

		    end_trace_message(buf);
		    down(&this->ad_odriver_sem);
		    if (dest->td_state != TD_ERROR) {
			if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data,OO_KERNEL_BUFFER) != buf->ptr - buf->data)
	    		    dest->td_state = TD_ERROR;
    		    }
		    up(&this->ad_odriver_sem);
		    reset_buffer(buf);
		    if (ii < last_tick)
			add_op(this, dest, buf, pre_op((i)));
		}
	    }
	}
	release_buffer(buf);
out:
	return;
}
#else
void aggregate_stop_trace(struct assembly_driver *this, struct trace_destination *dest)
{
	struct buffer *buf = get_buffer();
	field_t i;
	if (!buf)
	    goto out;
	add_op(this, dest, buf, MSG_STOP);
	for (i = 0; i < OP_MAX; i++) {
		add_tv(buf, ARG_MSG_ID, i);
		add_tv(buf, ARG_ULONG, pd(this)->opcounts[i]);
	}
	end_trace_message(buf);
	down(&this->ad_odriver_sem);
	if (dest->td_state != TD_ERROR) {
	    if (dest->td_od->od_ops->write(dest->td_od, buf->data, buf->ptr - buf->data) != buf->ptr - buf->data)
	        dest->td_state = TD_ERROR;
        }
	up(&this->ad_odriver_sem);
	release_buffer(buf);
out:
	return;
}
#endif

static void aggregate_create_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, int mode, void **context)
{
    PRE_HEADER(OP_CREATE);
    PRE_FOOTER;
}
static void aggregate_create_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, int mode, int ret, void **context)
{
    POST_HEADER(OP_CREATE);
    POST_FOOTER;
}

static void aggregate_lookup_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_LOOKUP);
    PRE_FOOTER;
}
static void aggregate_lookup_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, dentry_t *ret_dentry, void **context)
{
	POST_HEADER(OP_LOOKUP);
	POST_FOOTER;
}

static void aggregate_unlink_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_UNLINK);
    PRE_FOOTER;
}

static void aggregate_unlink_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, int ret, void **context)
{
    POST_HEADER(OP_UNLINK);
    POST_FOOTER;
}
static void aggregate_symlink_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, const char *symname, void **context)
{
    PRE_HEADER(OP_SYMLINK);
    PRE_FOOTER;
}

static void aggregate_symlink_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *dir, dentry_t *dentry, const char *symname, int ret, void **context)
{
    POST_HEADER(OP_SYMLINK);
    POST_FOOTER;
}

static void aggregate_link_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry, void **context)
{
    PRE_HEADER(OP_LINK);
    PRE_FOOTER;
}
static void aggregate_link_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry, int ret, void **context)
{
    POST_HEADER(OP_LINK);
    POST_FOOTER;
}

static void aggregate_mkdir_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode* dir, struct dentry *dentry, int mode, void **context)
{
    PRE_HEADER(OP_MKDIR);
    PRE_FOOTER;
}
static void aggregate_mkdir_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, int ret, void **context)
{
    POST_HEADER(OP_MKDIR);
    POST_FOOTER;
}
static void aggregate_rmdir_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, void **context)
{
    PRE_HEADER(OP_RMDIR);
    PRE_FOOTER;
}
static void aggregate_rmdir_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int ret, void **context)
{
    POST_HEADER(OP_RMDIR);
    POST_FOOTER;
}
static void aggregate_mknod_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, unsigned long dev, void **context)
{
    PRE_HEADER(OP_MKNOD);
    PRE_FOOTER;
}
static void aggregate_mknod_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, unsigned long dev, int ret, void **context)
{
    POST_HEADER(OP_MKNOD);
    POST_FOOTER;
}
static void aggregate_rename_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, void **context)
{
    PRE_HEADER(OP_RENAME);
    PRE_FOOTER;
}
static void aggregate_rename_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, int ret, void **context)
{
    POST_HEADER(OP_RENAME);
    POST_FOOTER;
}
static void aggregate_readlink_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, char *buffer, int bufsiz, void **context)
{
    PRE_HEADER(OP_READLINK);
    PRE_FOOTER;
}
static void aggregate_readlink_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, char *buffer, int bufsiz, int ret, void **context)
{
    POST_HEADER(OP_READLINK);
    POST_FOOTER;
}
static void aggregate_follow_link_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct nameidata *nd, void **context)
{
    PRE_HEADER(OP_FOLLOW_LINK);
    PRE_FOOTER;
}
static void aggregate_follow_link_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct nameidata *nd, int ret, void **context)
{
    POST_HEADER(OP_FOLLOW_LINK);
    POST_FOOTER;
}
static void aggregate_permission_pre(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode, int mask, void **context)
{
    PRE_HEADER(OP_PERMISSION);
    PRE_FOOTER;
}
static void aggregate_permission_post(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode, int mask, int ret, void **context)
{
    POST_HEADER(OP_PERMISSION);
    POST_FOOTER;
}
static void aggregate_inode_revalidate_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, void **context)
{
    PRE_HEADER(OP_INODE_REVALIDATE);
    PRE_FOOTER;
}
static void aggregate_inode_revalidate_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, int ret, void **context)
{
    POST_HEADER(OP_INODE_REVALIDATE);
    POST_FOOTER;
}
static void aggregate_setattr_pre(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct iattr *ia, void **context)
{
    PRE_HEADER(OP_SETATTR);
    PRE_FOOTER;
}
static void aggregate_setattr_post(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct iattr *ia, int ret, void **context)
{
    POST_HEADER(OP_SETATTR);
    POST_FOOTER;
}
/* file ops */
static void aggregate_read_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, char *buffer, size_t count, loff_t *ppos, void **context)
{
    PRE_HEADER(OP_READ);
    PRE_FOOTER;
}
static void aggregate_read_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, char *buffer, size_t count, loff_t *ppos, ssize_t ret, void **context)
{
    POST_HEADER(OP_READ);
    POST_FOOTER;
}
static void aggregate_write_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, const char *buffer, size_t count, loff_t *ppos, void **context)
{
    PRE_HEADER(OP_WRITE);
    PRE_FOOTER;
}
static void aggregate_write_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, const char *buffer, size_t count, loff_t *ppos, ssize_t ret, void **context)
{
    POST_HEADER(OP_WRITE);
    POST_FOOTER;
}
static void aggregate_readdir_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void *dirent, filldir_t filldir, int count, void **context)
{
    PRE_HEADER(OP_READDIR);
    PRE_FOOTER;
}
static void aggregate_readdir_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void *firent, filldir_t filldir, int count, int ret, void **context)
{
    POST_HEADER(OP_READDIR);
    POST_FOOTER;
}
static void aggregate_poll_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, poll_table *wait, void **context)
{
    PRE_HEADER(OP_POLL);
    PRE_FOOTER;
}
static void aggregate_poll_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, poll_table *wait, int ret, void **context)
{
    POST_HEADER(OP_POLL);
    POST_FOOTER;
}
static void aggregate_ioctl_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg, void **context)
{
    PRE_HEADER(OP_IOCTL);
    PRE_FOOTER;
}
static void aggregate_ioctl_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg, int ret, void **context)
{
    POST_HEADER(OP_IOCTL);
    POST_FOOTER;
}
static void aggregate_mmap_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, vm_area_t *vma, void **context)
{
    PRE_HEADER(OP_MMAP);
    PRE_FOOTER;
}
static void aggregate_mmap_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, vm_area_t *vma, int ret, void **context)
{
    POST_HEADER(OP_MMAP);
    POST_FOOTER;
}
static void aggregate_flush_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void **context)
{
    PRE_HEADER(OP_FLUSH);
    PRE_FOOTER;
}
static void aggregate_flush_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int ret, void **context)
{
    POST_HEADER(OP_FLUSH);
    POST_FOOTER;
}
static void aggregate_release_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, void **context)
{
    PRE_HEADER(OP_RELEASE);
    PRE_FOOTER;
}
static void aggregate_release_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, int ret, void **context)
{
    POST_HEADER(OP_RELEASE);
    POST_FOOTER;
}
static void aggregate_fsync_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, dentry_t *dentry, int datasync, void **context)
{
    PRE_HEADER(OP_FSYNC);
    PRE_FOOTER;
}
static void aggregate_fsync_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, dentry_t *dentry, int datasync, int ret, void **context)
{
    POST_HEADER(OP_FSYNC);
    POST_FOOTER;
}
static void aggregate_fasync_pre(struct assembly_driver *this, struct trace_destination *dest, int fd, file_t *file, int flag, void **context)
{
    PRE_HEADER(OP_FASYNC);
    PRE_FOOTER;
}
static void aggregate_fasync_post(struct assembly_driver *this, struct trace_destination *dest, int fd, file_t *file, int flag, int ret, void **context)
{
    POST_HEADER(OP_FASYNC);
    POST_FOOTER;
}
static void aggregate_lock_pre(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int cmd, struct file_lock *fl, void **context)
{
    PRE_HEADER(OP_LOCK);
    PRE_FOOTER;
}
static void aggregate_lock_post(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int cmd, struct file_lock *fl, int ret, void **context)
{
    POST_HEADER(OP_LOCK);
    POST_FOOTER;
}
static void aggregate_read_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER(OP_READ_INODE);
    PRE_FOOTER;
}
static void aggregate_read_inode_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    POST_HEADER(OP_READ_INODE);
    POST_FOOTER;
}
static void aggregate_write_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, int sync, void **context)
{
    PRE_HEADER(OP_WRITE_INODE);
    PRE_FOOTER;
}
static void aggregate_write_inode_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, int sync, void **context)
{
    POST_HEADER(OP_WRITE_INODE);
    POST_FOOTER;
}
static void aggregate_put_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER(OP_PUT_INODE);
    PRE_FOOTER;
}
static void aggregate_put_inode_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    POST_HEADER(OP_PUT_INODE);
    POST_FOOTER;
}
static void aggregate_delete_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER(OP_DELETE_INODE);
    PRE_FOOTER;
}
static void aggregate_delete_inode_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    POST_HEADER(OP_DELETE_INODE);
    POST_FOOTER;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void aggregate_statfs_pre(struct assembly_driver *this, struct trace_destination *dest, struct statfs *buffer, void **context)
#else
static void aggregate_statfs_pre(struct assembly_driver *this, struct trace_destination *dest, struct kstatfs *buffer, void **context)
#endif
{
    PRE_HEADER(OP_STATFS);
    PRE_FOOTER;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void aggregate_statfs_post(struct assembly_driver *this, struct trace_destination *dest, struct statfs *buffer, int ret, void **context)
#else
static void aggregate_statfs_post(struct assembly_driver *this, struct trace_destination *dest, struct kstatfs *buffer, int ret, void **context)
#endif
{
    POST_HEADER(OP_STATFS);
    POST_FOOTER;
}
static void aggregate_clear_inode_pre(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    PRE_HEADER(OP_CLEAR_INODE);
    PRE_FOOTER;
}
static void aggregate_clear_inode_post(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context)
{
    POST_HEADER(OP_CLEAR_INODE);
    POST_FOOTER;
}
static void aggregate_d_revalidate_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int flags, void **context)
{
    PRE_HEADER(OP_D_REVALIDATE);
    PRE_FOOTER;
}
static void aggregate_d_revalidate_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int flags, int ret, void **context)
{
    POST_HEADER(OP_D_REVALIDATE);
    POST_FOOTER;
}
static void aggregate_d_hash_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *name, void **context)
{
    PRE_HEADER(OP_D_HASH);
    PRE_FOOTER;
}
static void aggregate_d_hash_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *name, int ret, void **context)
{
    POST_HEADER(OP_D_HASH);
    POST_FOOTER;
}
static void aggregate_d_compare_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *a, qstr_t *b, void **context)
{
    PRE_HEADER(OP_D_COMPARE);
    PRE_FOOTER;
}
static void aggregate_d_compare_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *a, qstr_t *b, int ret, void **context)
{
    POST_HEADER(OP_D_COMPARE);
    POST_FOOTER;
}
static void aggregate_d_delete_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_D_DELETE);
    PRE_FOOTER;
}
static void aggregate_d_delete_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int ret, void **context)
{
    POST_HEADER(OP_D_DELETE);
    POST_FOOTER;
}

static void aggregate_d_release_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context)
{
    PRE_HEADER(OP_D_RELEASE);
    PRE_FOOTER;
}
static void aggregate_d_release_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context)
{
    POST_HEADER(OP_D_RELEASE);
    POST_FOOTER;
}
static void aggregate_d_iput_pre(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, inode_t *inode, void **context)
{
    PRE_HEADER(OP_D_IPUT);
    PRE_FOOTER;
}
static void aggregate_d_iput_post(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, inode_t *inode, void **context)
{
    POST_HEADER(OP_D_IPUT);
    POST_FOOTER;
}

static void aggregate_readpage_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, void **context)
{
    PRE_HEADER(OP_READPAGE);
    PRE_FOOTER;
}

static void aggregate_readpage_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, int ret, void **context)
{
    POST_HEADER(OP_READPAGE);
    POST_FOOTER;
}

static void aggregate_writepage_pre(struct assembly_driver *this, struct trace_destination *dest, struct page *page, void **context)
{
    PRE_HEADER(OP_WRITEPAGE);
    PRE_FOOTER;
}

static void aggregate_writepage_post(struct assembly_driver *this, struct trace_destination *dest, struct page *page, int ret, void **context)
{
    POST_HEADER(OP_WRITEPAGE);
    POST_FOOTER;
}

static void aggregate_prepare_write_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, void **context)
{
        PRE_HEADER(OP_PREPARE_WRITE);
        PRE_FOOTER;
}

void aggregate_prepare_write_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, int ret, void **context)
{
        POST_HEADER(OP_PREPARE_WRITE);
        POST_FOOTER;
}

void aggregate_commit_write_pre(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, void **context)
{
        PRE_HEADER(OP_COMMIT_WRITE);
        PRE_FOOTER;
}

void aggregate_commit_write_post(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, int ret, void **context)
{
        POST_HEADER(OP_COMMIT_WRITE);
	POST_FOOTER;
}


static struct assembly_ops aggregate_ad_ops = {
init: aggregate_init,
release: aggregate_release,
ao_llseek_pre: aggregate_llseek_pre,
ao_llseek_post: aggregate_llseek_post,
ao_create_pre: aggregate_create_pre,
ao_create_post: aggregate_create_post,
ao_lookup_pre: aggregate_lookup_pre,
ao_lookup_post: aggregate_lookup_post,
ao_unlink_pre: aggregate_unlink_pre,
ao_unlink_post: aggregate_unlink_post,
ao_symlink_pre: aggregate_symlink_pre,
ao_symlink_post: aggregate_symlink_post,
ao_link_pre: aggregate_link_pre,
ao_link_post: aggregate_link_post,
ao_mkdir_pre: aggregate_mkdir_pre,
ao_mkdir_post: aggregate_mkdir_post,
ao_inode_revalidate_pre: aggregate_inode_revalidate_pre,
ao_inode_revalidate_post: aggregate_inode_revalidate_post,
ao_permission_pre: aggregate_permission_pre,
ao_permission_post: aggregate_permission_post,
ao_follow_link_pre: aggregate_follow_link_pre,
ao_follow_link_post: aggregate_follow_link_post,
ao_readlink_pre: aggregate_readlink_pre,
ao_readlink_post: aggregate_readlink_post,
ao_rename_pre: aggregate_rename_pre,
ao_rename_post: aggregate_rename_post,
ao_mknod_pre: aggregate_mknod_pre,
ao_mknod_post: aggregate_mknod_post,
ao_rmdir_pre: aggregate_rmdir_pre,
ao_rmdir_post: aggregate_rmdir_post,
ao_setattr_pre: aggregate_setattr_pre,
ao_setattr_post: aggregate_setattr_post,
ao_read_pre : aggregate_read_pre,
ao_read_post : aggregate_read_post,
ao_write_pre : aggregate_write_pre,
ao_write_post : aggregate_write_post,
ao_readdir_pre : aggregate_readdir_pre,
ao_readdir_post : aggregate_readdir_post,
ao_poll_pre : aggregate_poll_pre,
ao_poll_post : aggregate_poll_post,
ao_ioctl_pre : aggregate_ioctl_pre,
ao_ioctl_post : aggregate_ioctl_post,
ao_mmap_pre : aggregate_mmap_pre,
ao_mmap_post : aggregate_mmap_post,
ao_open_pre: aggregate_open_pre,
ao_open_post: aggregate_open_post,
ao_flush_pre : aggregate_flush_pre,
ao_flush_post : aggregate_flush_post,
ao_release_pre : aggregate_release_pre,
ao_release_post : aggregate_release_post,
ao_fsync_pre : aggregate_fsync_pre,
ao_fsync_post : aggregate_fsync_post,
ao_fasync_pre : aggregate_fasync_pre,
ao_fasync_post : aggregate_fasync_post,
ao_lock_pre : aggregate_lock_pre,
ao_lock_post : aggregate_lock_post,
/* super ops */
ao_read_inode_pre : aggregate_read_inode_pre,
ao_read_inode_post : aggregate_read_inode_post,
ao_write_inode_pre : aggregate_write_inode_pre,
ao_write_inode_post : aggregate_write_inode_post,
ao_put_inode_pre : aggregate_put_inode_pre,
ao_put_inode_post : aggregate_put_inode_post,
ao_delete_inode_pre : aggregate_delete_inode_pre,
ao_delete_inode_post : aggregate_delete_inode_post,
ao_statfs_pre : aggregate_statfs_pre,
ao_statfs_post : aggregate_statfs_post,
ao_clear_inode_pre : aggregate_clear_inode_pre,
ao_clear_inode_post : aggregate_clear_inode_post,
/* dentry ops */
ao_d_revalidate_pre : aggregate_d_revalidate_pre,
ao_d_revalidate_post : aggregate_d_revalidate_post,
ao_d_hash_pre : aggregate_d_hash_pre,
ao_d_hash_post : aggregate_d_hash_post,
ao_d_compare_pre : aggregate_d_compare_pre,
ao_d_compare_post : aggregate_d_compare_post,
ao_d_delete_pre : aggregate_d_delete_pre,
ao_d_delete_post : aggregate_d_delete_post,
ao_d_release_pre : aggregate_d_release_pre,
ao_d_release_post : aggregate_d_release_post,
ao_d_iput_pre : aggregate_d_iput_pre,
ao_d_iput_post : aggregate_d_iput_post,

ao_readpage_pre: aggregate_readpage_pre,
ao_readpage_post: aggregate_readpage_post,
ao_writepage_pre: aggregate_writepage_pre,
ao_writepage_post: aggregate_writepage_post,
ao_prepare_write_pre: aggregate_prepare_write_pre,
ao_prepare_write_post: aggregate_prepare_write_post,
ao_commit_write_pre: aggregate_commit_write_pre,
ao_commit_write_post: aggregate_commit_write_post,

ao_start_trace: aggregate_start_trace,
ao_stop_trace: aggregate_stop_trace
};

struct assembly_driver *create_aggregate_driver()
{
    struct assembly_driver *tmp = (struct assembly_driver *)KMALLOC(sizeof(struct assembly_driver), GFP_KERNEL);

    if (tmp) {
	tmp->ad_type = AD_AGGREGATE;
	tmp->ad_ops = &aggregate_ad_ops;
	tmp->ad_next = NULL;
	tmp->ad_private_data = NULL;
    }
    return tmp;


}


void destroy_aggregate_driver(struct assembly_driver *aggregate)
{
    ASSERT(aggregate);
    ASSERT(aggregate->ad_type == AD_AGGREGATE);
    ASSERT(aggregate->ad_private_data == NULL);

    /* only free as of now */
    KFREE(aggregate);
}
