/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
#ifndef __COMMON_H
#define __COMMON_H

#include "messages.h"
#include "structures.h"
/*
 * common.h
 * This file includes the data structures and constants shared between the kernel and user space
 * This essentially defines the ioctl structures
 */

#define OD_COMPRESS	0
#define OD_CKSUM	1
#define OD_ENCRYPT	2
#define OD_FILE		3
#define OD_SOCKET	4
#define OD_NETLINK	5
//#define OD_RAW		6
#define OD_BUFFER       6
#define OD_ASYNC        7

#define AD_STREAM	8	/* Assembly driver to turn operations into a stream. */
#define AD_AGGREGATE    9       /* Assembly driver for aggregate counters */

struct od_file_params {
	char *name; /* pathname of file to trace to. NULL terminated, user level */
	int buf_size; /* buffer size, 0 == no buffering */
	int fd;       /* file descriptor-based file driver, name must be NULL */
	int sync_writes;     /* should use O_SYNC writes */
};

struct od_async_params {
	char *name; /* pathname of file to trace to. NULL terminated, user level */
	int buf_size; /* buffer size */
	int buf_count; /* number of buffers to use */
	int fd;       /* file descriptor-based file driver, name must be NULL */
	int sync_writes;     /* should use O_SYNC writes */
};


struct od_buffer_params {
	int buf_size; /* buffer size, 0 == buffer of lower level */
};

#define ENCRYPT_BLOWFISH 1
struct od_encrypt_params {
	char *key; /* key for encryption, user space */
        int keylen; /* length of key */
	int algorithm; /* algorithm to use */
};

struct od_compress_params {
	int in_buf_size;
};

/* struct for HMAC MD5 in checksum driver */
struct od_cksum_params {
	char *key; /* key for encryption, user space */
	int keylen;
};

/* verbosity options for stream assembly driver */
#define STR_POST_OP              0x1
#define STR_ARGS                 0x2
#define STR_PID                  0x4
#define STR_PNAME                0x8
#define STR_UID                  0x10
#define STR_GID                  0x20
#define STR_SYSCALL_ID           0x40
#define STR_TIMESTAMP            0x80
#define STR_UTIMESTAMP           0x100
#define STR_DATA                 0x200
#define STR_DATA_CKSUM           0x400
#define STR_DATA_WITHOUT_READ    0x800

#ifndef __KERNEL__
#define MAKE_VERBOSITY(M) { M, #M }
#ifdef _VERBOSITY_CONSTANTS_
struct {
    int id;
    char *name;
} _verbosity_opts[] = {
   MAKE_VERBOSITY(STR_POST_OP),
   MAKE_VERBOSITY(STR_ARGS),
   MAKE_VERBOSITY(STR_PID),
   MAKE_VERBOSITY(STR_PNAME),
   MAKE_VERBOSITY(STR_UID),
   MAKE_VERBOSITY(STR_GID),
   MAKE_VERBOSITY(STR_SYSCALL_ID),
   MAKE_VERBOSITY(STR_TIMESTAMP),
   MAKE_VERBOSITY(STR_UTIMESTAMP),
   MAKE_VERBOSITY(STR_DATA),
   MAKE_VERBOSITY(STR_DATA_CKSUM),
   MAKE_VERBOSITY(STR_DATA_WITHOUT_READ),
   { 0, NULL }
};
#endif
#undef MAKE_VERBOSITY
#endif
struct ad_stream_params {
	int ad_verbosity;
};


struct user_output_driver {
	int od_type;	/* Type of destination. OD_* */
	union {
		struct od_file_params     od_file;
		struct od_async_params    od_async;
		struct od_compress_params od_compress;
		struct od_buffer_params   od_buffer;
		struct od_encrypt_params  od_encrypt;
		struct od_cksum_params    od_cksum;
	} u;
};

struct user_assembly_driver {
	int ad_type; 	/* Type of assembly driver. AD_ */
	union {
		struct ad_stream_params ad_stream;
	} u;
};
struct user_tracecond {
	enum node_type tc_type;
	int tc_follow_true;
	int tc_follow_false;
	union {
		char tc_map[BITMAP_SIZE]; /* bitmap for TC_OP or TC_SYSCALL */
		unsigned tc_ival;    /* for TC_xUID, TC_xGID, TC_PID, TC_SID, TC_INO */
		struct time tc_time; /* for TC_TIME */
		char *tc_name;     /* for TC_PNAME, TC_NAME and TC_EXT */
	} u;
};

/* Steps:
 * 1. Create output driver stack.  Pass in an array of ints OD_* and their private data.  (Optional)
 * 2. Create assembly driver stack. Pass in output driver id.
 * 3. Send rules to kernel.  Pass assembly drive id.
 */
/* Corresponding teardown:
 * 3. Rule
 * 2 & 1. Assembly, Output Driver
 * 0. Destroy everything for unmount.
 */

#define IOCBASE 10
/* Add output driver stack. Returns integer ID. */
#define TRACEFS_IOCTL_ADD_OUTPUT_DRIVERS	_IOW(0x15, IOCBASE + 0, int)
/* Add assembly driver on top of output driver stack. */
#define TRACEFS_IOCTL_ADD_ASSEMBLY_DRIVERS	_IOW(0x15, IOCBASE + 1, int)
/* Add rule to assembly/output driver stack. */
#define TRACEFS_IOCTL_ADD_RULE			_IOW(0x15, IOCBASE + 2, int)
/* Free output driver stack. */
#define TRACEFS_IOCTL_DEL_OUTPUT_DRIVERS	_IOW(0x15, IOCBASE + 3, int)
/* Free assembly driver stack. */
#define TRACEFS_IOCTL_DEL_ASSEMBLY_DRIVERS	_IOW(0x15, IOCBASE + 4, int)
/* Free rule on top of assembly driver stack. */
#define TRACEFS_IOCTL_DEL_RULE			_IOW(0x15, IOCBASE + 5, int)
/* Free all rules, output drivers, and assembly drivers. */
#define TRACEFS_IOCTL_TEARDOWN			_IOW(0x15, IOCBASE + 6, int)
/* Start the trace */
#define TRACEFS_IOCTL_START_TRACE               _IOW(0x15, IOCBASE + 7, int)
/* Stop the trace */
#define TRACEFS_IOCTL_STOP_TRACE                _IOW(0x15, IOCBASE + 8, int)

#endif /* __COMMON_H */
