/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include "fist.h"
#include "namecache.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
#define HASH_BITS 6
#define HASH_MASK ((1 << HASH_BITS) - 1)
#define NUM_BUCKETS (1 << HASH_BITS) /* for the time being */

/**
 * Initializes the name cache.
 * Creates the first level hash table
 *
 * @return new name cache if successful, NULL otherwise
 */
struct namecache *namecache_init()
{
    int i;
    struct namecache *namecache = KMALLOC(sizeof(struct namecache) + sizeof(struct list_head) * NUM_BUCKETS, GFP_KERNEL);
    struct list_head *head;
    print_entry_location();
    if (namecache) {
	head = namecache->nc_list;
	i = NUM_BUCKETS;
	do {
	    INIT_LIST_HEAD(head);
	    head++;
	    i--;
	} while (i);
    }
    spin_lock_init(&namecache->nc_lock);
    print_exit_location();
    return namecache;
}

/**
 * Hash value of the pattern.
 * @param pattern to hash
 * @param whether hashing extension or the full name
 * @return hash value
 */
static inline long hash(struct qstr *pattern, int ext)
{
	int i = pattern->len;
	unsigned char *c = (unsigned char *)(pattern->name + i - 1);
	unsigned long hash = 0;
	while(i--) {
		hash += *c;
#ifdef TRACE_EXT
		if (ext && *c == '.')
			break;
#endif
		c--;

	}
	return hash & HASH_MASK;
}

/**
 * Allocates a new pattern.
 * @param New pattern to create
 * @return New <code>struct name</code> for the pattern, NULL on failure
 */
static struct name *allocate_pattern(struct qstr *pattern, int ext)
{
	int i;
	struct list_head *head;
	struct name *name = KMALLOC(sizeof(struct name) + NUM_BUCKETS * sizeof(struct list_head), GFP_KERNEL);
	int len = pattern->len;
	int dot_index = 0;
#ifdef TRACE_EXT
	if (ext) {
		dot_index = len;
		while (--dot_index > 0 && pattern->name[dot_index] != '.')
		;
		/* dot index is zero if there is no extension */
		len -= dot_index;
	}
#endif

	if (!name)
		goto out_err;
	name->n_pat.name = KMALLOC(len, GFP_KERNEL);
	if (!name->n_pat.name)
		goto out_free;
	atomic_set(&name->n_refcount, 1);
	spin_lock_init(&name->n_lock);
	memcpy((void *)name->n_pat.name, pattern->name + dot_index, len);
	name->n_pat.len = len;
	INIT_LIST_HEAD(&name->n_list);
	i = name->n_buckets = NUM_BUCKETS;
	head = name->n_fid;
	do {
		INIT_LIST_HEAD(head);
		head++;
		i--;
	} while (i);
	return name;
out_free:
	KFREE(name);
out_err:
	name = NULL;
	return name;

}

/**
 * Frees the specified <code>name</code> structure.
 * Also frees the inner <code>fid</code> objects
 */
static void free_pattern(struct name **name)
{
	struct name *pat = *name;
	struct fid *fid;
	struct list_head *tmp;
	int i = pat->n_buckets;
	struct list_head *head = pat->n_fid;
	do {
		tmp = head->next;
		while (tmp != head) {
			struct list_head *next = tmp->next;
			fid = list_entry(tmp, struct fid, f_list);
			list_del(&fid->f_list);
			KFREE(fid);
			tmp = next;
		}
		head++;
		i--;
	} while(i);
	KFREE((char *)pat->n_pat.name);
	KFREE(*name);
	*name = NULL;
}

/**
 * Releases the name cache.
 * Also releases any internal structures.
 * Disregards the reference counts, should only be called
 * during unmount.
 * @param Name cache to release
 */

void namecache_release(struct namecache *namecache)
{
	int i = NUM_BUCKETS - 1;
	struct list_head *head;
	struct list_head *tmp;
	struct name *name;
	print_entry_location();
	spin_lock(&namecache->nc_lock);
	head = namecache->nc_list;
	while (i) {
		tmp = head->next;
		while (tmp != head) {
			struct list_head *next = tmp->next;
			name = list_entry(tmp, struct name, n_list);
			list_del(&name->n_list);
			free_pattern(&name);
			tmp = next;
		}
		head++;
		i--;
	}
	spin_unlock(&namecache->nc_lock);
	KFREE(namecache);
	print_exit_location();

}


#ifdef TRACE_EXT
/*
 * needle is the pattern ".c"
 * haystack is the full name "foo.c"
 */
static inline int extension_matches(struct qstr *needle, struct qstr *haystack)
{
	int i, j;
	if (needle->len <= haystack->len) {
		for (i = haystack->len - 1, j = needle->len - 1; j >= 0; i--, j--) {
			if (haystack->name[i] != needle->name[j])
				return 0;
		}
		return 1;
	}
	return 0;
}
#endif

/**
 * Finds the specified pattern.
 * To be used when adding a new pattern.
 * @param bucket to search in
 * @param pattern to search
 * @param ext 1 if extension based, 0 if full name
 * @return matched <code>name</code> structure.
 */
static struct name *find_pattern(struct list_head *head, struct qstr *pattern, int ext)
{
	struct list_head *tmp = head;
	struct name *match;
	for (;;) {
		tmp = tmp->next;
		match = NULL;
		if (tmp == head)
			break;
		match = list_entry(tmp, struct name, n_list);
#ifdef TRACE_EXT
		if (ext) {
			if (extension_matches(&match->n_pat, pattern))
				break;
		} else
#endif
			if (match->n_pat.len == pattern->len && strncmp(match->n_pat.name, pattern->name, pattern->len) == 0)
				break;

	}
	return match;
}

static inline struct name *lookup_pattern(struct namecache *namecache, struct qstr *pattern, int ext)
{
    struct list_head *head = namecache->nc_list + hash(pattern, ext);
    return find_pattern(head, pattern, ext);
}

/**
 * Adds a new pattern node in the hash table.
 * @param name cache to add new pattern to
 * @param pattern to add
 * @param ext 1 if extension based, 0 if full name
 * @return new name node, or refcount incremented existing node
 */
struct name *add_pattern(struct namecache *namecache, struct qstr *pattern, int ext)
{
    struct list_head *head;
    struct name *match;
    print_entry_location();
    spin_lock(&namecache->nc_lock);
    head = namecache->nc_list + hash(pattern, ext);
    match = find_pattern(head, pattern, ext);
    if (match) {
	atomic_inc(&match->n_refcount);
    } else if ((match = allocate_pattern(pattern, ext)) != NULL) {
	list_add(&match->n_list, head);
    }
    spin_unlock(&namecache->nc_lock);
    print_exit_status(match);
    return match;
}

static void _unlocked_release_pattern(struct name *pattern)
{
    int i;
    struct fid *fid;
    if (atomic_dec_and_test(&pattern->n_refcount)) {
	for (i = pattern->n_buckets - 1; i >= 0; i--) {
	    struct list_head *tmp = pattern->n_fid[i].next;
	    while (tmp != &pattern->n_fid[i]) {
		struct list_head *next = tmp->next;
		fid = list_entry(tmp, struct fid, f_list);
		list_del(&fid->f_list);
		KFREE(list_entry(tmp, struct fid, f_list));
		tmp = next;
	    }
	}
	list_del(&pattern->n_list);
	free_pattern(&pattern);
    }
}

void release_pattern(struct namecache *namecache, struct name *pattern)
{
    print_entry_location();
    spin_lock(&namecache->nc_lock);
    _unlocked_release_pattern(pattern);
    spin_unlock(&namecache->nc_lock);
    print_exit_location();

}

/**
 * Hash of inode.
 * @param inode number
 * @param kdev object
 * @return hash value
 */
static inline unsigned long hash_fid(unsigned long ino, kdev_t *kdev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	return (ino + HASHDEV(*kdev)) & HASH_MASK;
#else
	return (ino + *kdev) & HASH_MASK;
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
}



/**
 * Looks up inode for specified pattern.
 * @param pattern to look up for
 * @param inode number
 * @param kdev of inode
 * @return FID node if found, NULL otherwise
 */
static inline struct fid *_unlocked_lookup_fid(struct name *table, unsigned long ino, kdev_t *kdev)
{
	struct list_head *head = table->n_fid + hash_fid(ino, kdev);
	struct list_head *tmp = head;
	struct fid *fid;
	for (;;) {
		tmp = tmp->next;
		fid = NULL;
		if (tmp == head)
			break;
		fid = list_entry(tmp, struct fid, f_list);
		if (fid->f_ino == ino && (fid->f_dev == *kdev))
			break;
	}
	return fid;
}

struct fid *lookup_fid(struct name *table, unsigned long ino, kdev_t *kdev)
{
    struct fid *fid;
    spin_lock(&table->n_lock);
    fid = _unlocked_lookup_fid(table, ino, kdev);
    spin_unlock(&table->n_lock);
    return fid;
}

/**
 * Add new inode to pattern.
 * @param pattern to add inode to
 * @param inode to add
 * @return 1 if fid got added, 0 otherwise
 */
static int add_fid(struct name *table, struct inode *ino)
{
    struct list_head *head;
    struct fid *fid;
    print_entry_location();
    spin_lock(&table->n_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    head = table->n_fid + hash_fid(ino->i_ino, &ino->i_dev);
#else
    head = table->n_fid + hash_fid(ino->i_ino, &ino->i_sb->s_dev);
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    fid = KMALLOC(sizeof(struct fid), GFP_KERNEL);
    if (fid) {
	INIT_LIST_HEAD(&fid->f_list);
	fid->f_ino = ino->i_ino;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	fid->f_dev = ino->i_dev;
#else
	fid->f_dev = ino->i_sb->s_dev;
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	list_add(&fid->f_list, head);
    }
    spin_unlock(&table->n_lock);
    print_exit_status(fid != NULL);
    return (fid != NULL);
}

/**
 * Add inode to name cache
 * @param name cache to add to.
 * @param dentry of the inode
 * @param inode to add
 * @return 1 if inode was added, 0 otherwise
 */

int add_inode(struct namecache *namecache, struct dentry *dentry, struct inode *inode)
{
	/* for now, use full file names */
	int ret = 0;
	struct name *name;
	print_entry_location();
	spin_lock(&namecache->nc_lock);
	name = lookup_pattern(namecache, &dentry->d_name, 0);
	if (name)
		ret = add_fid(name, inode);
#ifdef TRACE_EXT
	name = lookup_pattern(namecache, &dentry->d_name, 1);
	if (name)
		ret |= add_fid(name, inode);
#endif
	spin_unlock(&namecache->nc_lock);
	print_exit_status(ret);

	return ret;
}

/**
 * This is called for unlink
 */


static inline void _unlocked_delete_inode(struct namecache *namecache, struct dentry *dentry, struct inode *inode, int ext)
{
    struct name *name;
    struct fid *fid;
    print_entry_location();
    name = lookup_pattern(namecache, &dentry->d_name, ext);
    if (name) {
	spin_lock(&name->n_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	fid = _unlocked_lookup_fid(name, inode->i_ino, &inode->i_dev);
#else
	fid = _unlocked_lookup_fid(name, inode->i_ino, &inode->i_sb->s_dev);
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (fid) {
	    list_del(&fid->f_list);
	    KFREE(fid);
	}
	spin_unlock(&name->n_lock);
    }

    print_exit_location();

}

void delete_inode(struct namecache *namecache, struct dentry *dentry, struct inode *inode)
{
    spin_lock(&namecache->nc_lock);

    _unlocked_delete_inode(namecache, dentry, inode, 0);
#ifdef TRACE_EXT
    _unlocked_delete_inode(namecache, dentry, inode, 1);
#endif
    spin_unlock(&namecache->nc_lock);
}


/*
 * Removes all instances of the inode in the name patterns
 * This is a potentially expensive operation if there are a large
 * number of patterns in the system.
 */

void delete_inode_global(struct namecache *namecache, struct inode *inode)
{
    int i;
    struct fid *fid;
    print_entry_location();
    spin_lock(&namecache->nc_lock);
    for (i = 0; i < NUM_BUCKETS; i++) {
	struct list_head *head = &namecache->nc_list[i];
	struct name *match;
	struct list_head *tmp = head;
	for (;;) {
		tmp = tmp->next;
		match = NULL;
		if (tmp == head)
			break;
		match = list_entry(tmp, struct name, n_list);
		spin_lock(&match->n_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		fid = _unlocked_lookup_fid(match, inode->i_ino, &inode->i_dev);
#else
		fid = _unlocked_lookup_fid(match, inode->i_ino, &inode->i_sb->s_dev);
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		if (fid) {
		    list_del(&fid->f_list);
		    KFREE(fid);
		}
		spin_unlock(&match->n_lock);
	}

    }
    spin_unlock(&namecache->nc_lock);


    print_exit_location();
}

/*
 * Rename operation
 * need to move the inode from old name to new name
 * Also, need to remove the new name's inode (if any)
 */
static void _namecache_rename(struct namecache *namecache, struct dentry *old_dentry, struct dentry *new_dentry, int ext)
{
    struct name *dst_pattern;
    struct name *src_pattern;
    struct fid *src_fid = NULL;
    struct fid *dst_fid = NULL;
    print_entry_location();
    //spin_lock(&namecache->nc_lock);  //BUG, already locked by caller

    src_pattern = lookup_pattern(namecache, &old_dentry->d_name, ext);
    dst_pattern = lookup_pattern(namecache, &new_dentry->d_name, ext);
    if (src_pattern)
	spin_lock(&src_pattern->n_lock);
    if (dst_pattern && dst_pattern != src_pattern)
	spin_lock(&dst_pattern->n_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (src_pattern &&
	(src_fid = _unlocked_lookup_fid(src_pattern, old_dentry->d_inode->i_ino, &old_dentry->d_inode->i_dev)) != NULL) {
#else
    if (src_pattern &&
	(src_fid = _unlocked_lookup_fid(src_pattern, old_dentry->d_inode->i_ino, &old_dentry->d_inode->i_sb->s_dev)) != NULL) {
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	list_del(&src_fid->f_list);
	KFREE(src_fid);
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (dst_pattern && new_dentry->d_inode &&
	(dst_fid = _unlocked_lookup_fid(dst_pattern, new_dentry->d_inode->i_ino, &new_dentry->d_inode->i_dev)) != NULL) {
#else
    if (dst_pattern && new_dentry->d_inode &&
	(dst_fid = _unlocked_lookup_fid(dst_pattern, new_dentry->d_inode->i_ino, &new_dentry->d_inode->i_sb->s_dev)) != NULL) {
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	list_del(&dst_fid->f_list);
	KFREE(dst_fid);
    }
    if (dst_pattern) {
	    dst_fid = KMALLOC(sizeof(struct fid), GFP_KERNEL);
	    if (dst_fid) {
		struct list_head *head;
		INIT_LIST_HEAD(&dst_fid->f_list);
		dst_fid->f_ino = old_dentry->d_inode->i_ino;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		dst_fid->f_dev = old_dentry->d_inode->i_dev;
#else
		dst_fid->f_dev = old_dentry->d_inode->i_sb->s_dev;
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		head = dst_pattern->n_fid + hash_fid(dst_fid->f_ino, &dst_fid->f_dev);
		list_add(&dst_fid->f_list, head);
	    } else {
		fist_dprint(2, "namecache_rename: Error! out of memory\n");
	    }
    }

    if (dst_pattern && dst_pattern != src_pattern)
	spin_unlock(&dst_pattern->n_lock);
    if (src_pattern)
	spin_unlock(&src_pattern->n_lock);
    //spin_unlock(&namecache->nc_lock); //BUG, caller will unlock

    print_exit_location();

}

void namecache_rename(struct namecache *namecache, struct dentry *old_dentry, struct dentry *new_dentry)
{
	spin_lock(&namecache->nc_lock);

	_namecache_rename(namecache, old_dentry, new_dentry, 0);
#ifdef TRACE_EXT
	_namecache_rename(namecache, old_dentry, new_dentry, 1);
#endif
	spin_unlock(&namecache->nc_lock);

}


