/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include "gengraph.h"
#include "common.h"

extern int optind;

extern int yydebug;

static void dfs_print(struct block *root, char *indent, int level);
static int dfs_const(struct block *root, int *index, struct user_tracecond **array, char *indent, int level);
static void print_output_drivers(struct user_output_driver *list, int count);
static void print_assembly_drivers(struct user_assembly_driver *list, int count);

void usage(const char *name) {
	fprintf(stderr, "%s [OPTIONS] MOUNTPOINT\n", name);
	fprintf(stderr, "Valid options are:\n");
	fprintf(stderr, "-p\tPrint out parsed values.\n");
	exit(1);
}

int main(int argc, char *argv[])
{
    struct graphset *gs;
    struct graphset *cgs;
    char indent[80];
    struct user_tracecond *tc = NULL;
    int idx = 0;
    int i;
    int c;
    int print = 0;
    char *mount;
    int fd;
    int ret;


    while ((c = getopt(argc, argv, "p")) != -1) {
	switch(c) {
    	case 'p':
		print = 1;
		break;
    	default:
		usage(argv[0]);
	}
    }

    if (argc - optind != 1) {
	usage(argv[0]);
    }

    mount = argv[optind];


    cgs = gs = tracefs_compile();

    memset(indent, ' ', sizeof(indent));
    indent[0] = '\0';

    while (cgs) {
	int destid;
	void *data;
	void *d;

	if (print) {
		printf("Printing ...\n");
		dfs_print(cgs->root, indent, 0);
		print_output_drivers(cgs->ods, cgs->n_ods);
		print_assembly_drivers(cgs->ads, cgs->n_ads);
	}

	printf("Constructing ...\n");

	if(cgs->root) dfs_const(cgs->root, &idx, &tc, indent, 0);

	if (print) {
		for (i = 0; i < idx; i++) {
	    		printf("%d: ", i);
	    		switch(tc[i].tc_type) {
	    		case TC_TRUE:
				printf("TRUE");
				break;
	    		case TC_FALSE:
				printf("FALSE");
				break;
	    		case TC_CUID:
	    		case TC_CGID:
	    		case TC_FUID:
	    		case TC_FGID:
	    		case TC_PID:
	    		case TC_PGRP:
	    		case TC_SID:
	    		case TC_INO:
				printf("%s = %d", get_node_type(tc[i].tc_type), tc[i].u.tc_ival);
				break;
	    		case TC_PNAME:
				printf("%s = %s", get_node_type(tc[i].tc_type), tc[i].u.tc_map);
				break;
	    		case TC_NAME:
	    		case TC_EXT:
				printf("%s = %s", get_node_type(tc[i].tc_type), tc[i].u.tc_name);
				break;
			case TC_SYSCALL:
			case TC_OP: {
			    int j;
			    printf("%s = ", get_node_type(tc[i].tc_type));
			    for (j = 0; j < BITMAP_SIZE; j++)
				printf("0x%x ", tc[i].u.tc_map[j]);
			    break;
			}


	    		default:
				printf("ERR: Invalid Type: %d\n", tc[i].tc_type);
				exit(1);
	    		}
	    		printf(" [ TRUE -> %d ] [ FALSE -> %d ]", tc[i].tc_follow_true, tc[i].tc_follow_false);
	    		printf("\n");
		}
	}

	/* Stuff Output Drivers */
	/* Stuff Assembly Drivers + Destid */
	/* Stuff Rule + Destid */

	data = malloc((cgs->n_ods * sizeof(struct user_output_driver)) + sizeof(unsigned int));
	*((unsigned int *)data) = cgs->n_ods;
	d = data + sizeof(unsigned int);
	memcpy(d, cgs->ods, cgs->n_ods * sizeof(struct user_output_driver));

	fd = open(mount, O_RDONLY);
	if (fd == -1) {
		perror("open");
		exit(1);
	}

	destid = ioctl(fd, TRACEFS_IOCTL_ADD_OUTPUT_DRIVERS, data);
	if (destid == -1) {
		perror("ioctl(TRACEFS_IOCTL_ADD_OUTPUT_DRIVERS)");
		exit(1);
	}
	printf("Destination ID: %d\n", destid);

	free(data);

	data = malloc((cgs->n_ods * sizeof(struct user_assembly_driver)) + (2 *sizeof(unsigned int)));
	*((unsigned int *)data) = destid;
	*(((unsigned int *)data) + 1) = cgs->n_ads;
	d = data + (2 * sizeof(unsigned int));
	memcpy(d, cgs->ads, cgs->n_ads * sizeof(struct user_assembly_driver));

	ret = ioctl(fd, TRACEFS_IOCTL_ADD_ASSEMBLY_DRIVERS, data);
	if (ret == -1) {
		perror("ioctl(TRACEFS_IOCTL_ADD_ASSEMBLY_DRIVERS");
		exit(1);
	}

	free(data);

	data = malloc((idx * sizeof(struct user_tracecond)) + (2 * sizeof(unsigned int)));
	*((unsigned int *)data) = destid;
	*(((unsigned int *)data) + 1) = idx;
	d = data + (2 * sizeof(unsigned int));
	memcpy(d, tc, idx * sizeof(struct user_tracecond));

	ret = ioctl(fd, TRACEFS_IOCTL_ADD_RULE, data);
	if (ret == -1) {
		perror("ioctl(TRACEFS_IOCTL_ADD_RULE");
		exit(1);
	}

	printf("All done A-OK!\n");

	free(data);

	cgs = cgs->next;
    }

    tracefs_compile_done();

    return 0;
}

static void print_output_drivers(struct user_output_driver *tmp, int count)
{
	while (count--) {
		printf("Type = %d\n", tmp->od_type);
		switch(tmp->od_type) {
			case OD_FILE:
				printf("   File = %s\n", tmp->u.od_file.name);
				break;
			case OD_COMPRESS:
				printf("   Compress\n");
				break;
			case OD_CKSUM:
				printf("   Checksum\n");
				break;
			case OD_ENCRYPT:
				printf("   Encrypt\n");
				break;
			case OD_ASYNC:
				printf("   Async\n");
				break;
			default:
				printf("   Unknown\n");
		}
		tmp++;
	}
}

static void print_assembly_drivers(struct user_assembly_driver *tmp, int count)
{
	while (count--) {
		printf("Type = %d\n", tmp->ad_type);
		switch(tmp->ad_type) {
			case AD_STREAM:
				break;
			case AD_AGGREGATE:
				break;
			default:
				printf(" Unknown\n");
		}
		tmp++;
	}
}
static void dfs_print(struct block *root, char *indent, int level) {
	int skip = 0;
	if (!root)
		return;

	indent[level] = ' ';
	indent[level + 1] = '\0';
	switch(root->type) {
	case TC_TRUE:
		skip = 1;
		printf("%s->TRUE-<\n", indent);
		break;
	case TC_FALSE:
		skip = 1;
		printf("%s->FALSE-<\n", indent);
		break;
	case TC_PNAME:
	case TC_NAME:
	case TC_EXT:
	    printf("%s%s: %s\n", indent, get_node_type(root->type), root->u.svalue);
	    break;
	default:
		printf("%s%s: %d\n", indent, get_node_type(root->type), root->u.ivalue);
		break;
	}
	if (!skip) {
		printf("%sTRUE\n", indent);
		if (root->et.succ) {
			dfs_print(root->et.succ, indent, level + 1);
		}
		printf("%sFALSE\n", indent);
		if (root->ef.succ) {
			dfs_print(root->ef.succ, indent, level + 1);
		}
	}
	indent[level] = '\0';
	indent[level + 1] = ' ';
}

int dfs_const(struct block *root, int *index, struct user_tracecond **array, char *indent, int level) {
	int myindex = *index;

	root->index = myindex;

	indent[level] = ' ';
	indent[level + 1] = '\0';

	(*index)++;

	*array = (struct user_tracecond *) realloc(*array, sizeof(struct user_tracecond) * (myindex + 1));
	(*array)[myindex].tc_type = root->type;
	switch(root->type) {
	case TC_TRUE:
	//	printf("%s->TRUE-<\n", indent);
		break;
	case TC_FALSE:
	//	printf("%s->FALSE-<\n", indent);
		break;
	case TC_CUID:
	case TC_CGID:
	case TC_FUID:
	case TC_FGID:
	case TC_PID:
	case TC_PGRP:
	case TC_SID:
	case TC_INO:
	//	printf("%sUID: %d\n", indent, root->u.ivalue);
		(*array)[myindex].u.tc_ival = root->u.ivalue;
		break;
	case TC_PNAME:
		strncpy((*array)[myindex].u.tc_map, root->u.svalue, sizeof((*array)[myindex].u.tc_map));
		break;
	case TC_NAME:
	case TC_EXT:
	    (*array)[myindex].u.tc_name = root->u.svalue;
	    break;
	case TC_SYSCALL:
	case TC_OP:
	    memcpy((*array)[myindex].u.tc_map, root->u.svalue, sizeof((*array)[myindex].u.tc_map));
	break;

	default:
		printf("ERR: Invalid Type: %d\n", root->type);
		exit(1);
	}
	//printf("%sTRUE\n", indent);
	if (root->et.succ) {
		if (root->et.succ->index == -1) {

			(*array)[myindex].tc_follow_true = dfs_const(root->et.succ, index, array, indent, level + 1);
		} else {
			(*array)[myindex].tc_follow_true = root->et.succ->index;
		}
	} else {
		(*array)[myindex].tc_follow_true = -1;
	}
	//printf("%sFALSE\n", indent);
	if (root->ef.succ) {
		if (root->ef.succ->index == -1) {

			(*array)[myindex].tc_follow_false = dfs_const(root->ef.succ, index, array, indent, level + 1);
		} else {
			(*array)[myindex].tc_follow_false = root->ef.succ->index;
		}
	} else {
		(*array)[myindex].tc_follow_false = -1;
	}
	indent[level] = '\0';
	indent[level + 1] = ' ';


	return myindex;
}
